<?php
  class MyTemplate   
  {
    public    $fn = null;
    protected $templateBody = '';
    protected $vars = array();
    public    $body = '';
    
    private   $subReports = array();
    private   $subReportsBands = array();    
    private   $subReportsParameters = array();
      
    private   $subReportsCounts = array();
    private   $subReportsSums = array();

    private   $subReportsGlobalCounts = array();
    private   $subReportsGlobalSums = array();

    private   $subReportsFieldsToSum = array();

    private   $sums = array();
    private   $counts = array();

    public    $clearUnparsedTokens = true;


    function __construct ($fn = null)
    {
      set_magic_quotes_runtime(false);      // we do not need this!!!
      $this->fn = $fn;
      if (!is_null($this->fn))
        if (!$this->loadTemplateFile($fn))
          throw new Exception("Cannot open template file $fn (or file is empty).");
        else
          {
            // we got the template
            $this->extractSubReports();
          }         
        
    }

    public function loadTemplateFile($fn)
    {      
      if ((!$fn) )
      {
       throw new Exception("Cannot open template file $fn"); 
      }

      if (file_exists($fn))
        $this->templateBody = @file_get_contents($fn);      
      else if (file_exists(getcwd()."/".$fn))
        $this->templateBody = @file_get_contents(getcwd()."/".$fn);      
      else if (file_exists(getcwd()."/tpl/".$fn))
        $this->templateBody = @file_get_contents(getcwd()."/tpl/".$fn);          
      else if (file_exists(getcwd()."/templates/".$fn))
        $this->templateBody = @file_get_contents(getcwd()."/templates/".$fn);         
      else return false;
      
      if ((!$this->templateBody) or (strlen($this->templateBody) < 1) )
        return false;
      else
        return true;
    }    

    private function extractSubReports()
    {
      $tokens = array();

      if (preg_match_all('/<!-- report name="(\\w+)"( params\\((.*?)\\))? begin -->(.*)<!-- report name="(\\1)" end -->/sim',$this->templateBody, $tokens) <= 0)
        return;

      $this->subReports = array();
      $this->subReportsBands = array();

      if ( (is_array($tokens)) && (isset($tokens[1])) && (sizeof($tokens[1])>0) )
        foreach ($tokens[1] as $k => $v)
        {
          $this->subReports[strtolower($v)] = $tokens[0][$k];
          $this->extractBandsForSubReport($v);
          if (isset($tokens[3][$k]))
            $this->extractAdditionalSubReportParameters($v,$tokens[3][$k]);  
          $this->extractFieldsToWatchForSubReport($v);
        }        

    }

    private function extractFieldsToWatchForSubReport($name)
    {
      if (!isset($this->subReports[$name]))
        throw new Exception("Tried to extract FIELDS TO WATCH for subreport that do not exist ($name).");
    
      $this->subReportsFieldsToSum[$name] = array();
      
      if (preg_match_all('/\\(%(\\w+)\\.(\\w+)%\\)/sim', $this->subReports[$name], $tokens)>0)
      {
        if ( (is_array($tokens)) && (isset($tokens[1])) && (sizeof($tokens[1])>0) )
        {
          foreach ($tokens[1] as $k => $v)
          {
            if (strtolower($v) == "avg")
              $this->subReportsFieldsToSum[$name][] = $tokens[2][$k];
            else if (strtolower($v) == "sum")
              $this->subReportsFieldsToSum[$name][] = $tokens[2][$k];
          }
          $this->subReportsFieldsToSum[$name] = array_unique($this->subReportsFieldsToSum[$name]);
        }
      }
    }
    
    
    
    private function extractAdditionalSubReportParameters($name, $rawtext)
    {
      if (!isset($this->subReports[$name]))
        throw new Exception("Tried to extract ADDITIONAL PARAMETERS of subreport that do not exist ($name).");
    
      $this->subReportsParameters[$name] = array();
      
      if (trim($rawtext) == '')
        return;
      
      $opts = array();
      if (preg_match_all('/(\\w+=\\"\\w+\\")+/sim',$rawtext, $opts) <= 0)
        return;      
      if ( (is_array($opts)) && (isset($opts[0])) )
        foreach ($opts[0] as $v)
          {
            $o = explode("=",$v);
            $this->subReportsParameters[$name][$o[0]] = substr($o[1],1,-1);                
          }                     
    }
         
  
    private function extractBandsForSubReport($name)
    {
      $name = strtolower($name);
      
      $tokens = array();
      if (preg_match_all('/<!-- (\\w+) begin -->(.*)<!-- (\\1) end -->/sim',$this->subReports[$name], $tokens) <= 0)
        return;
      
      $this->subReportsBands[$name] = array();      
      
      if ( (is_array($tokens)) && (isset($tokens[3])) && (sizeof($tokens[3])>0) )
        foreach ($tokens[3] as $k => $v)
          $this->subReportsBands[strtolower($name)][$v] = $tokens[2][$k];        
    
    }  
    
  
  public function hasSubReports()
  {
    return (is_array($this->subReports) && (sizeof($this->subReports) > 0) );
  }
  
  private function getSubReportsHeaderBand($name)
  {
    if (!isset($this->subReports[$name]))
      throw new Exception("Tried to open HEADER band of subreport that do not exist ($name).");
      
    if (isset($this->subReportsBands[$name]["header"]))
      return ($this->subReportsBands[$name]["header"]);
    else
    {
      // here add some logic to determine what is header if no bands are marked in subreport
      return '';
    }
  }

  private function getSubReportsDetailBand($name)
  {
    if (!isset($this->subReports[$name]))
      throw new Exception("Tried to open DETAIL band of subreport that do not exist ($name).");
      
    if (isset($this->subReportsBands[$name]["detail"]))
      return ($this->subReportsBands[$name]["detail"]);
    else
    {
      // no detail band, so if there is no header and no footer then, all is detail
      if ( (!isset($this->subReportsBands[$name]["header"])) && (!isset($this->subReportsBands[$name]["footer"])) )
        return ($this->subReports[$name]);
      else
        return '';
    }
  }

  private function getSubReportsDetailEmptyBand($name)
  {
    if (!isset($this->subReports[$name]))
      throw new Exception("Tried to open DETAIL_EMPTY band of subreport that do not exist ($name).");
      
    if (isset($this->subReportsBands[$name]["detail_empty"]))
      return ($this->subReportsBands[$name]["detail_empty"]);
    else
    {
      // no detail band, so if there is no header and no footer then, all is detail
      if ( (!isset($this->subReportsBands[$name]["header"])) && (!isset($this->subReportsBands[$name]["footer"])) )
        return ($this->subReports[$name]);
      else
        return '';
    }
  }  
  
  
  private function getSubReportsFooterBand($name)
  {
    if (!isset($this->subReports[$name]))
      throw new Exception("Tried to open FOOTER band of subreport that do not exist ($name).");
      
    if (isset($this->subReportsBands[$name]["footer"]))
      return ($this->subReportsBands[$name]["footer"]);
    else
    {
      // here add some logic to determine what is header if no bands are marked in subreport
      return '';
    }  
  }

  
  private function getSubReportParameter($subReportName,$parameterName)
  {
    if (isset($this->subReportsParameters[$subReportName][$parameterName]))
      return ($this->subReportsParameters[$subReportName][$parameterName]);
    else
      return null;
  }


  private function replaceAllTokensInText(&$text, $subReportName = null, $arrayIndex = null)
  {
    $tokens = array();
    if ( preg_match_all('/\\(%([\\w|:|\\.]+)%\\)/sim', $text, $tokens) > 0 )
      {
        foreach ($tokens[1] as $v)
            $this->replaceOneTokenPlaceholderWithValue($text,$v,$subReportName, $arrayIndex);
      }

  }

  private function clearAllEmptyTokensInText(&$text)
  {
    $tokens = array();
    if ( preg_match_all('/\\(%([\\w|:|\\.]+)%\\)/sim', $text, $tokens) > 0 )
      {
        foreach ($tokens[1] as $v)
            $this->replaceOneTokenPlaceholderWithValue($text,$v,array());
      }

  }


  private function applyTokenFilterToTokenValue($tokenValue, $filterName)
  {    
    switch (strtolower($filterName)) 
    {
       case "raw":
        $new = $tokenValue;
         break;
       case "upper":
        $new = strtoupper($tokenValue);
         break;
       case "lower":
        $new = strtolower($tokenValue);
         break;
       case "ucwords":
        $new = ucwords($tokenValue);
         break;         
       case "nohtml":
         $new = htmlentities($tokenValue,ENT_QUOTES, 'UTF-8');
         break;
       default:
        $new = ($tokenValue);
         break;
    }    
    return $new;    
  }

  private function replaceOneTokenPlaceholderWithValue(&$text, $tokenname, $subReportName = null, $arrayIndex = null) 
  {   
    if (is_null($subReportName))
      $varsArray = $this->vars;
    else
    {
      $varsArray = $this->vars[$subReportName][$arrayIndex];
    }
       
    if (strpos($tokenname,':') !== false)
     { 
       $realtokenname = substr($tokenname,0,strpos($tokenname,':'));       
       $tokenfilter = substr($tokenname,(strpos($tokenname,':')+1)); 
       if (isset($varsArray[$realtokenname]))       
        $new = $this->applyTokenFilterToTokenValue($varsArray[$realtokenname],$tokenfilter);
       else
        $new = '';
        // replace the token with filtered value
        $text = str_ireplace("(%".$tokenname."%)",$new,$text);              
     }
    else if (strpos($tokenname,'.') !== false)
     { 
       $realtokenname = strtolower(trim(substr($tokenname,strpos($tokenname,'.')+1)));       
       $tokenoperation = strtolower(trim(substr($tokenname,0,(strpos($tokenname,'.'))))); 
      
       if ($tokenoperation == "sum")
       {
        $new = $this->subReportsSums[$subReportName][$realtokenname];
       }
       else if ($tokenoperation == "count")
       {
        $new = $this->subReportsCounts[$subReportName];
       }
       else if ($tokenoperation == "avg")
       {
        $new = ($this->subReportsSums[$subReportName][$realtokenname] / $this->subReportsCounts[$subReportName] );   
       }
       
        // replace the token with calculated value
        $text = str_ireplace("(%".$tokenname."%)",$new,$text);              
     }     
     else
      {
        // no modifier, just replace the value
        if (isset($varsArray[$tokenname]))
          $text = str_ireplace("(%".$tokenname."%)",$varsArray[$tokenname],$text);                      
        else
          $text = str_ireplace("(%".$tokenname."%)",'',$text);        
      }
              
  }
  
  private function updateSubReportsWatches($subReportName, $currentArrayIndex)
  {
    isset($this->subReportsCounts[$subReportName]) ? $this->subReportsCounts[$subReportName]++ : $this->subReportsCounts[$subReportName] = 1;
    isset($this->subReportsGlobalCounts[$subReportName]) ? $this->subReportsGlobalCounts[$subReportName]++ : $this->subReportsGlobalCounts[$subReportName] = 1;

    if (count($this->subReportsFieldsToSum[$subReportName]) > 0)
      foreach ($this->subReportsFieldsToSum[$subReportName] as $fn)
        {
          isset($this->subReportsSums[$subReportName]) && isset($this->subReportsSums[$subReportName][$fn]) && isset($this->vars[$subReportName][$currentArrayIndex]) && isset($this->vars[$subReportName][$currentArrayIndex][$fn]) ? $this->subReportsSums[$subReportName][$fn] += $this->vars[$subReportName][$currentArrayIndex][$fn] : $this->subReportsSums[$subReportName][$fn] = $this->vars[$subReportName][$currentArrayIndex][$fn];
          isset($this->subReportsGlobalSums[$subReportName][$fn]) && isset($this->vars[$subReportName][$currentArrayIndex]) ? $this->subReportsGlobalSums[$subReportName][$fn] += $this->vars[$subReportName][$currentArrayIndex][$fn] : $this->subReportsGlobalSums[$subReportName][$fn] = $this->vars[$subReportName][$currentArrayIndex][$fn];        
        }

  }

  private function resetSubReportsTemporaryWatches($name)
  {
    $this->subReportsSums[$name] = array();
    $this->subReportsCounts[$name] = 0;
  }
 
 
  private function parseSingleSubReport($name)
  {
    if (!isset($this->subReports[$name]))
      return ;
    
    $breakColumnName = $this->getSubReportParameter($name,"break");
    $breakingThisSubReport = ((!is_null($breakColumnName)) && isset($this->vars[$name][0][$breakColumnName])) ? true : false;        
    $result = '';

    if ( (isset($this->vars[$name])) && (is_array($this->vars[$name])) && (sizeof($this->vars[$name])>0) )
    {

      if ($breakingThisSubReport)
        $previousKey = $this->vars[$name][0][$breakColumnName];    

      $rowcount = sizeof($this->vars[$name]);
      for ($i = 0;  $i < $rowcount; $i++)
      {
        $v = $this->vars[$name][$i];
        $firstTime = ($i == 0);
        
        if ( ($firstTime) || ( $breakingThisSubReport && ($previousKey != $v[$breakColumnName])) )
        {
          $result .= $this->getSubReportsHeaderBand($name);
          $this->resetSubReportsTemporaryWatches($name);
        }
        
        $result .= $this->getSubReportsDetailBand($name);

        if ( ( ($i >= ($rowcount-1)) ) || ( ($breakingThisSubReport) && ($v[$breakColumnName] != $this->vars[$name][$i+1][$breakColumnName]) ) )
          $result .= $this->getSubReportsFooterBand($name);           
        
        if ($breakingThisSubReport)            
          $previousKey =  $v[$breakColumnName];
        
        $this->updateSubReportsWatches($name, $i);
        
        $this->processAllTemplateLogicConstructsInText($result,$name,$i);

        $this->replaceAllTokensInText($result, $name, $i);
      }
        
    }
    else
    {
      $result .= $this->getSubReportsHeaderBand($name);       
      $result .= $this->getSubReportsDetailEmptyBand($name);
      $result .= $this->getSubReportsFooterBand($name);        
    }
    

    
    $this->body = preg_replace('/<!-- report name="'.$name.'"(.*)? begin -->(.*)<!-- report name="'.$name.'" end -->/sim',$result, $this->body);
  }
  
  private function extractIfParts($text)
  {
    $parts = array("logic" => null, "a" => null, "b"=>null, "operation" => null, "true" => null, "false" => null, "body" => "", "type" => "if");
    
    if (preg_match('/<!-- if (.*?) -->(.*?)<!-- endif -->/sim', $text, $tokens) < 1 )
      return $parts;
      
    if (sizeof($tokens)>0)
    {
      $parts['logic']   = $tokens[1];      
      $parts['body']    = $tokens[2];
    }
    else
      return $parts;
    
    $tokens = array();
    if  ((preg_match_all('/\\s?([^ ]+)/sim', $parts['logic'], $tokens) < 1) || (sizeof($tokens[0]) < 5) )
      {
        $parts['logic']   = null;
        return $parts;
      }
    else
      {
        $parts['a'] = trim($tokens[0][1]);
        $parts['b'] = trim($tokens[0][3]);        
        $parts['operation'] = trim($tokens[0][2]);                  
      }        
          
    if (stripos($parts['body'],'<!-- else -->') !== false)
    {
      // we have else part
      $tokens = '';
      $parts['true'] = substr($parts['body'],0,stripos($parts['body'],'<!-- else -->'));
      $parts['false'] = substr($parts['body'],stripos($parts['body'],'<!-- else -->')+strlen("<!-- else -->"));      
    }
    else
    {
      $parts['true'] = '';
    }
      
    return $parts;
  }
  
  private function getLogicalVariableValue($name, $subReportName = null, $subReportArrayIndex = 0)
  {
    if ($name[0] == "$")
      {
        $realName = substr($name,1);
        if (!is_null($subReportName))
          {
            return (isset($this->vars[$subReportName][$subReportArrayIndex][$realName]) ? $this->vars[$subReportName][$subReportArrayIndex][$realName] : null);
          }     
        else
          return isset($this->vars[$realName]) ? ($this->vars[$realName]) : null ;
      
      }
    else
      return $name;
  }
  
  private function evaluateOperation($parameters, $subReportName = null, $subReportArrayIndex = 0)
  {
    $a = $this->getLogicalVariableValue($parameters['a'],$subReportName,$subReportArrayIndex);
    $b = $this->getLogicalVariableValue($parameters['b'],$subReportName,$subReportArrayIndex);    
      
    switch (strtolower($parameters['type'])) 
    {
       case "if":
         switch (strtolower($parameters['operation'])) 
         {
            case "==":
              return ($a == $b);
              break;
            case "===":
              return ($a === $b);
              break;
            case "!=":
              return ($a != $b);
              break;
            case "!==":
              return ($a !== $b);
              break;
            case "<":
              return ($a < $b);         
              break;
            case ">":
              return ($a > $b);         
              break;
            case "is":
            $t = ( strtolower($b) === "true" );
            $f = ( strtolower($b) === "false" );            
            $o = ( strtolower($b) === "odd" );            
            $e = ( strtolower($b) === "even" );            
            if ($t)
            {
              return ( (bool) $a);         
            }
            else if ($f)
            {
              return ( (bool) !$a);         
            }
            else if ($o)
            {
              return ($a % 2);
            }
            else if ($e)
            {
              return (!($a % 2));
            }
              break;              
            default:
              return null;
         }  
         break;
        default:
          return null;    
         break;
    }  
  }
  
  
  private function processIfLogicInText($text, $ifText, $subReportName = null, $subReportArrayIndex = 0)
  {
    $parts = $this->extractIfParts($ifText);
    $result = $this->evaluateOperation($parts,$subReportName, $subReportArrayIndex);
    $len = strlen($ifText);
    $pos = strpos($text,$ifText);        
    
    if (!is_null($result) && ($pos !== false) &&  ( ( $result == true) || ($result == false) )  )
      {
        $res = substr($text,0,$pos);
        if ($result) 
        {
          $res .= $parts['true'];
        }
        else
          $res .= $parts['false'];
        $res .= substr($text,($pos+$len));
        
        return $res;        
      }
    else
      return $text;           
  }
  
  private function processAllTemplateLogicConstructsInText(&$text, $subReportName = null, $subReportArrayIndex = 0)
  {
    
    $tokens = array();
    
    if ( preg_match_all('/<!-- if (.*?) endif -->/sim', $text, $tokens) > 0 )
      {
        if ( (is_array($tokens[0])) && (count($tokens[0])>0) )
         foreach ($tokens[0] as $k => $v)
         {
           $text = $this->processIfLogicInText($text,$v,$subReportName, $subReportArrayIndex);
         }
      }      
  }
  
  private function parseSubReports()
  {
    if (!$this->hasSubReports())
      return;
    
    foreach ($this->subReports as $name => $value)
    {
      $this->parseSingleSubReport($name);
    }
    
  }
  
  
  public function assign($name, $value)
  {
    $this->vars[$name] = $value;    
  }
  

  
  public function parse($andShow = true)
  {
    $this->body = $this->templateBody;
    
    $this->parseSubReports();
    
    $this->processAllTemplateLogicConstructsInText($this->body);
      
    $this->replaceAllTokensInText($this->body);

    $this->clearAllEmptyTokensInText($this->body);


    if ($andShow)
      echo $this->body;
  }

  
  } // class
?>